import datetime
import json
import mimetypes
import re
import tempfile
import urllib

from twisted.internet import defer
from twisted.internet.interfaces import IConsumer
from twisted.protocols import ftp
from zope.interface import implements

from .http import OAuthHTTPClient, HTTPError


epoch = datetime.datetime(1970, 1, 1, 0, 0, 0)
date_pat = re.compile(r'(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})Z')

def parse_date(datestr):
    match = date_pat.match(datestr)
    if match:
        date = datetime.datetime(
            int(match.group(1)), int(match.group(2)), int(match.group(3)),
            int(match.group(4)), int(match.group(5)), int(match.group(6)))
        delta = date - epoch
        return delta.days * 24 * 60 * 60 + delta.seconds
    else:
        print "Could not parse date %r" % (datestr,)
        return 0


class FileReader(object):
    implements(ftp.IReadFile)

    def __init__(self, producer):
        self.producer = producer

    def send(self, consumer):
        print 'FileReader.send:', consumer
        self.producer.startProducing(consumer)
        return self.producer.complete


class ContentConsumer(object):
    """A simple consumer that writes the contents to a file."""

    implements(IConsumer)

    def __init__(self):
        self.producer = None
        self.fp = tempfile.TemporaryFile()
        self.length = 0

    def registerProducer(self, producer, streaming):
        self.producer = producer
        assert streaming

    def unregisterProducer(self):
        self.producer = None

    def write(self, data):
        self.fp.write(data)
        self.length += len(data)


class FileWriter(object):

    implements(ftp.IWriteFile)

    def __init__(self, shell, content_path, content_type):
        self.shell = shell
        self.content_path = content_path
        self.content_type = content_type
        self.content_consumer = None
        self.deferred_response = None
        self.body_complete = defer.Deferred()

    def receive(self):
        print "FileWriter.receive()"
        assert self.content_consumer is None, "receive() called twice"
        # Since we need to know the length of the file before we can
        # use make the API call, we need to receive the entire
        # contents to a temporary file before sending it.
        self.content_consumer = ContentConsumer()
        return defer.succeed(self.content_consumer)

    @defer.inlineCallbacks
    def close(self):
        print "FileWriter.close()"
        assert self.content_consumer is not None, "receive() not called"
        # Create a body producer that sends the contents of the
        # temporary file to the API request.
        self.content_consumer.fp.seek(0)
        headers = {
            'Content-Type': self.content_type,
            'Content-Length': self.content_consumer.length
            }
        response = yield self.shell._json_request(
            "PUT", self.content_path, self.content_consumer.fp, headers)


class U1FTPShell(object):

    implements(ftp.IFTPShell)

    def __init__(self, reactor, oauth_consumer, oauth_token,
                 base_url, content_base_url=None):
        self.reactor = reactor
        self.client = OAuthHTTPClient(reactor, oauth_consumer, oauth_token)
        self.base_url = base_url
        self.content_base_url = content_base_url or base_url
        self._user_info = None

    def _make_full_url(self, url):
        # XXX: according to the docs, we need to redirect content
        # requests to this host name.
        if url.startswith('/content'):
            return self.content_base_url + url
        else:
            return self.base_url + url

    @defer.inlineCallbacks
    def _json_request(self, method, url, body=None, headers=None):
        """Perform an HTTP request, and return the decoded JSON body."""
        url = self._make_full_url(url)
        print method, url
        try:
            body = yield self.client.requestJson(method, url, headers, body)
        except HTTPError, exc:
            if exc.code >= 400:
                try:
                    body = json.loads(exc.body)
                except ValueError:
                    print "Could not parse HTTP error body:\n" + exc.body
                    raise ftp.PermissionDeniedError("Unknown error")
                message = body['error'].encode('utf-8')
                # Handle error responses
                if exc.code == 403:
                    raise ftp.PermissionDeniedError(message)
                elif exc.code == 404:
                    raise ftp.FileNotFoundError(message)
                elif exc.code == 409:
                    raise ftp.IsADirectoryError(message)
                else:
                    # Is there a better error code for generic errors?
                    raise ftp.PermissionDeniedError(message)
            else:
                raise
        defer.returnValue(body)

    @defer.inlineCallbacks
    def _get_user_info(self):
        if self._user_info is None:
            self._user_info = yield self._json_request('GET', '')
            #from pprint import pprint
            #pprint(self._user_info)
        defer.returnValue(self._user_info)

    def _create_volume_map(self, user_info):
        """Create a mapping of single path component names to volume paths."""
        volume_map = {}
        for path in ([user_info['root_node_path']] +
                     user_info['user_node_paths']):
            # Give a nice name to the U1MS folder.
            if path == '/~/.ubuntuone/Purchased from Ubuntu One':
                name = 'Purchased Music'
            else:
                name = path.strip('/~').replace('/', '_')
            volume_map[name] = path
        return volume_map

    @defer.inlineCallbacks
    def _make_file_path(self, path):
        # Since we encode the volume in the first path component, we
        # require at least one path component.
        if len(path) < 1:
            raise ftp.FileNotFoundError(path)
        user_info = yield self._get_user_info()
        # Treat the first path segment as a volume name via the volume map.
        volume_map = self._create_volume_map(user_info)
        if path[0] not in volume_map:
            raise ftp.FileNotFoundError(path)
        defer.returnValue(
            urllib.quote('/'.join([volume_map[path[0]]] + path[1:])))

    @defer.inlineCallbacks
    def _get_file_info(self, path, include_children=False):
        full_path = yield self._make_file_path(path)
        if include_children:
            full_path += '?include_children=true'
        file_info = yield self._json_request('GET', full_path)
        #from pprint import pprint
        #pprint(file_info)
        defer.returnValue(file_info)

    @defer.inlineCallbacks
    def makeDirectory(self, path):
        print 'makeDirectory:', path
        if len(path) < 2:
            # We only support creating directories within a volume.
            raise ftp.PermissionDeniedError(path)
        full_path = yield self._make_file_path(path)
        response = yield self._json_request(
            'PUT', full_path, json.dumps({'kind': 'directory'}),
            {'Content-Type': 'application/json'})

    @defer.inlineCallbacks
    def removeDirectory(self, path):
        print 'removeDirectory:', path
        if len(path) < 2:
            # We only support removing directories within a volume.
            raise ftp.PermissionDeniedError(path)
        file_info = yield self._get_file_info(path)
        if file_info['kind'] != 'directory':
            raise ftp.IsNotADirectoryError(path)
        if file_info['has_children']:
            raise ftp.PermissionDeniedError("%r has children" % path)
        response = yield self._json_request(
            'DELETE', urllib.quote(file_info['resource_path']))

    @defer.inlineCallbacks
    def removeFile(self, path):
        print 'removeFile:', path
        file_info = yield self._get_file_info(path)
        # XXX: we can delete whole trees with this command.  Should we
        # just do it?
        if file_info['kind'] != 'file':
            raise ftp.IsADirectoryError(path)
        response = yield self._json_request(
            'DELETE', urllib.quote(file_info['resource_path']))

    @defer.inlineCallbacks
    def rename(self, from_path, to_path):
        print 'rename:', from_path, to_path
        if len(from_path) < 2:
            raise ftp.PermissionDeniedError(from_path)
        if len(to_path) < 2 or from_path[0] != to_path[0]:
            raise ftp.PermissionDeniedError(
                "Can only perform renames within a volume")
        new_path = '/' + '/'.join(to_path[1:])
        full_path = yield self._make_file_path(from_path)
        response = yield self._json_request(
            'PUT', full_path, json.dumps({'path': new_path}),
            {'Content-Type': 'application/json'})

    @defer.inlineCallbacks
    def access(self, path):
        print 'access:', path
        # The top level path is always accessible.
        if len(path) == 0:
            return
        file_info = yield self._get_file_info(path)
        # We can only access directories.
        if file_info['kind'] != 'directory':
            raise ftp.IsNotADirectoryError(path)

    def _get_stat_info(self, file_info, keys):
        result = []
        for key in keys:
            if key == 'size':
                result.append(file_info.get('size') or 0)
            elif key == 'directory':
                result.append(file_info.get('kind') == 'directory')
            elif key == 'permissions':
                result.append(
                    0755 if file_info.get('kind') == 'directory' else 0644)
            elif key == 'hardlinks':
                result.append(1)
            elif key == 'modified':
                result.append(parse_date(file_info.get('when_changed', '')))
            elif key == 'owner':
                result.append('user')
            elif key == 'group':
                result.append('group')
        return result

    @defer.inlineCallbacks
    def stat(self, path, keys=()):
        print 'stat:', (path, keys)
        file_info = yield self._get_file_info(path)
        defer.returnValue(self._get_stat_info(file_info, keys))

    @defer.inlineCallbacks
    def list(self, path, keys=()):
        print 'list:', (path, keys)
        file_list = []
        if len(path) == 0:
            # Build the top level virtual volume map.
            user_info = yield self._get_user_info()
            volume_map = self._create_volume_map(user_info)
            attrs = {
                'size': 0,
                'directory': True,
                'permissions': 0755,
                'hardlinks': 1,
                'modified': 0,
                'owner': 'user',
                'group': 'group'
                }
            for name in sorted(volume_map.keys()):
                file_list.append(
                    (name.encode('utf-8'), [attrs[key] for key in keys]))
        else:
            file_info = yield self._get_file_info(path, include_children=True)
            if file_info['kind'] != 'directory':
                raise ftp.IsNotADirectoryError(path)
            for child_info in file_info.get('children', []):
                file_name = child_info['path'].split('/')[-1].encode('utf-8')
                file_list.append(
                    (file_name, self._get_stat_info(child_info, keys)))
        defer.returnValue(file_list)

    @defer.inlineCallbacks
    def openForReading(self, path):
        print 'openForReading:', path
        file_info = yield self._get_file_info(path)
        if file_info['kind'] != 'file':
            raise ftp.IsADirectoryError(path)
        # Now start reading the content ...
        producer = yield self.client.requestProducer(
            'GET', self._make_full_url(urllib.quote(
                    file_info['content_path'].encode('utf-8'))))
        defer.returnValue(FileReader(producer))

    @defer.inlineCallbacks
    def openForWriting(self, path):
        print 'openForWriting:', path
        if len(path) < 2:
            # We only support writing files within a volume.
            raise ftp.PermissionDeniedError(path)
        # Get information about the parent directory.
        parent_info = yield self._get_file_info(path[:-1])
        if parent_info['kind'] != 'directory':
            raise ftp.PermissionDeniedError("Parent directory doesn't exist")
        content_path = urllib.quote(parent_info['content_path'] + '/' +
                                    path[-1]).encode('utf-8')
        content_type = (mimetypes.guess_type(path[-1])[0] or
                        'application/octet-stream')
        defer.returnValue(FileWriter(self, content_path, content_type))
