import functools
import json
import urllib2

from oauth import oauth
from twisted.internet import defer, threads
from twisted.internet.interfaces import IPullProducer
from zope.interface import implements


def defer_to_thread(function):
    """A decorator that wraps function calls with deferToThread"""
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        return threads.deferToThread(function, *args, **kwargs)
    return wrapper


class HTTPError(Exception):
    """An exception representing a failed HTTP request."""

    def __init__(self, url, code, msg, headers, body):
        self.url = url
        self.code = code
        self.msg = msg
        self.headers = headers
        self.body = body

    def __str__(self):
        return 'HTTP Error %s: %s' % (self.code, self.msg)


class HTTPProducer(object):
    """A producer that reads data from a urllib2 response object."""
    implements(IPullProducer)

    buffer_size = 1 << 12

    def __init__(self, reactor, response):
        self.reactor = reactor
        self.response = response
        self.consumer = None
        self.complete = defer.Deferred()

    def _finish(self):
        if self.consumer:
            self.consumer.unregisterProducer()
            self.consumer = None
        self.complete.callback(None)

    def startProducing(self, consumer):
        self.consumer = consumer
        self.consumer.registerProducer(self, False)

    @defer_to_thread
    def resumeProducing(self):
        assert self.consumer is not None
        data = self.response.read(self.buffer_size)
        if not data:
            self.reactor.callFromThread(self._finish)
        else:
            print "Writing data"
            self.reactor.callFromThread(self.consumer.write, data)

    @defer_to_thread
    def stopProducing(self):
        print "HTTPProducer.stopProducing()"
        self.response.close()
        self.consumer = None


class MyRequest(urllib2.Request):
    """A urllib2 Request class that allows the method to be overriden."""

    _method = None

    def set_method(self, method):
        self._method = method.upper()

    def get_method(self):
        if self._method is not None:
            return self._method
        else:
            return urllib2.Request.get_method(self)


class HTTPClient(object):

    def __init__(self, reactor):
        self.reactor = reactor

    def _request(self, method, url, headers=None, body=None):
        if headers is None:
            headers = {}
        request = MyRequest(url, headers=headers, data=body)
        request.set_method(method)
        try:
            return urllib2.urlopen(request)
        except urllib2.HTTPError, exc:
            # We rewrap the exception, so that Twisted code can see
            # the response body for failed requests.
            if exc.fp is None:
                body = ''
            else:
                body = exc.fp.read()
            raise HTTPError(
                exc.url, exc.code, exc.msg, exc.hdrs, body)

    @defer_to_thread
    def requestText(self, method, url, headers=None, body=None):
        response = self._request(method, url, headers, body)
        return response.read()

    @defer_to_thread
    def requestJson(self, method, url, headers=None, body=None):
        response = self._request(method, url, headers, body)
        return json.load(response)

    @defer_to_thread
    def requestProducer(self, method, url, headers=None, body=None):
        response = self._request(method, url, headers, body)
        return HTTPProducer(self.reactor, response)


class OAuthHTTPClient(HTTPClient):

    def __init__(self, reactor, oauth_consumer, oauth_token):
        super(OAuthHTTPClient, self).__init__(reactor)
        self.oauth_consumer = oauth_consumer
        self.oauth_token = oauth_token

    def _request(self, method, url, headers=None, body=None):
        if headers is None:
            headers = {}
        oauth_request = oauth.OAuthRequest.from_consumer_and_token(
            self.oauth_consumer, self.oauth_token, method, url)
        signing_method = oauth.OAuthSignatureMethod_PLAINTEXT()
        oauth_request.sign_request(
            signing_method, self.oauth_consumer, self.oauth_token)
        headers.update(oauth_request.to_header())
        return super(OAuthHTTPClient, self)._request(method, url, headers, body)
