import base64
import json
import socket
import urllib

import keyring
from oauth import oauth
from twisted.cred import checkers, credentials, portal
from twisted.cred.error import UnauthorizedLogin
from twisted.internet import defer
from twisted.protocols import ftp
from zope.interface import implements

from .ftpshell import U1FTPShell
from .http import HTTPClient, OAuthHTTPClient, HTTPError


SERVICE_NAME = 'Ubuntu One FTP Server'


class U1CredentialsChecker(object):

    implements(checkers.ICredentialsChecker)

    credentialInterfaces = (credentials.IUsernamePassword,)

    def __init__(self, reactor, sso_base_url, ping_url):
        self.reactor = reactor
        self.sso_base_url = sso_base_url
        assert self.sso_base_url.endswith('/')
        self.ping_url = ping_url

    def get_hostname(self):
        return socket.gethostname()

    def _encode_basic_auth(self, credentials):
        """Build an HTTP Basic auth header from the given credentials."""
        auth_details = '%s:%s' % (credentials.username, credentials.password)
        return {
            'Authorization': 'Basic %s' % base64.b64encode(auth_details)}

    @defer.inlineCallbacks
    def _check_creds(self, credentials):
        client = HTTPClient(self.reactor)
        headers = self._encode_basic_auth(credentials)
        try:
            yield client.requestText(
                'GET', self.sso_base_url + 'authentications', headers=headers)
        except HTTPError:
            defer.returnValue(False)
        defer.returnValue(True)

    @defer.inlineCallbacks
    def _issue_token(self, credentials):
        # Issue the OAuth token through Ubuntu SSO
        client = HTTPClient(self.reactor)
        headers = self._encode_basic_auth(credentials)
        token_name = 'Ubuntu One @ %s [u1ftp]' % socket.gethostname()
        try:
            data = yield client.requestJson(
                'GET', self.sso_base_url + 'authentications?' +
                urllib.urlencode({'ws.op': 'authenticate',
                                  'token_name': token_name}),
                headers=headers)
        except HTTPError:
            raise UnauthorizedLogin("Could not issue a token")

        # Now ask Ubuntu One to pull the token into its token store
        consumer = oauth.OAuthConsumer(
            data['consumer_key'], data['consumer_secret'])
        token = oauth.OAuthToken(data['token'], data['token_secret'])
        client = OAuthHTTPClient(self.reactor, consumer, token)
        try:
            yield client.requestText('POST', self.ping_url)
        except HTTPError:
            raise UnauthorizedLogin("Could not transfer token to Ubuntu One")
        defer.returnValue((consumer, token))

    @defer.inlineCallbacks
    def requestAvatarId(self, credentials):
        print "U1CredentialsChecker.requestAvatarId(%s)" % credentials.username
        creds_ok = yield self._check_creds(credentials)
        if not creds_ok:
            print "Failed to log in."
            raise UnauthorizedLogin("Bad email address or password")

        # Now that we know the user entered the correct password,
        # retrieve the OAuth token from the keyring or issue a new
        # token.
        token_string = keyring.get_password(SERVICE_NAME, credentials.username)
        if token_string is None:
            print "Issuing a new OAuth token."
            consumer, token = yield self._issue_token(credentials)
            token_string = json.dumps(dict(
                    token=token.key,
                    token_secret=token.secret,
                    consumer_key=consumer.key,
                    consumer_secret=consumer.secret))
            keyring.set_password(
                SERVICE_NAME, credentials.username, token_string)
        defer.returnValue(token_string)


class U1FTPRealm(object):

    implements(portal.IRealm)

    def __init__(self, reactor, base_url, content_base_url=None):
        self.reactor = reactor
        self.base_url = base_url
        self.content_base_url = content_base_url

    def requestAvatar(self, avatarId, mind, *interfaces):
        for iface in interfaces:
            if iface is ftp.IFTPShell:
                token_data = json.loads(avatarId)
                oauth_consumer = oauth.OAuthConsumer(
                    token_data['consumer_key'], token_data['consumer_secret'])
                oauth_token = oauth.OAuthToken(
                    token_data['token'], token_data['token_secret'])
                avatar = U1FTPShell(
                    self.reactor, oauth_consumer, oauth_token,
                    self.base_url, self.content_base_url)
                return (ftp.IFTPShell, avatar,
                        getattr(avatar, 'logout', lambda: None))
        raise NotImplementedError("Only IFTPShell is supported by this realm")


def make_portal(reactor, base_url, content_base_url,
                sso_base_url, sso_ping_url):
    realm = U1FTPRealm(reactor, base_url, content_base_url)
    checker = U1CredentialsChecker(reactor, sso_base_url, sso_ping_url)
    return portal.Portal(realm, [checker])
